"""MOHID simulator module of the API."""
from typing import List, Optional

from inductiva import types, tasks, simulators

ALLOWD_MOHID_COMMANDS = ["MohidWater.exe", "MohidLand.exe"]


@simulators.simulator.mpi_enabled
class MOHID(simulators.Simulator):
    """Class to invoke a generic MOHID simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the MOHID simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "mohid"

    def run(self,
            input_dir: Optional[str],
            *,
            command: str = "MohidWater.exe",
            working_dir: str = "",
            on: types.ComputationalResources,
            run_ddc: bool = False,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            command: MOHID command to run. Can be `MohidWater.exe` or
                `MohidLand.exe`.
            working_dir: Path (relative to the input directory) to the directory
                where the simulation will run. The command picked and
                domain consolidation will run on that folder.
            run_ddc: Boolean that will determine if we are going to run 
                domain consolidation (`MohidDDC.exe`) after the
                simulation ends. This will combine the results of the multiple
                MPI processes into one.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """

        self._check_vcpus(n_vcpus, on)

        if command not in ALLOWD_MOHID_COMMANDS:
            raise ValueError(
                f"Command {command} not supported.\n"
                f"Available commands are: {ALLOWD_MOHID_COMMANDS}.")

        #We are using the intel MPI from oneapi
        #We don't have support for this MPI on the back end yet, this is why
        #we are not creating an MPICONFIG.
        mpi_command = "mpirun "
        if n_vcpus:
            mpi_command += f"-np {n_vcpus} "

        commands = [f"{mpi_command} /usr/local/bin/{command}"]

        if run_ddc:
            commands.append("/usr/local/bin/MohidDDC.exe")

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           run_subprocess_dir=working_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           **kwargs)



"""COAWST simulator module of the API."""

import os
from typing import List, Optional

from inductiva import types, tasks, simulators
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class COAWST(simulators.Simulator):
    """Class to invoke a generic COAWST simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the COAWST simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "coawst"

    def _validate_build_script(self, build_script: str) -> None:
        """
        Validates the build script by checking if required environment variables
        are correctly set.

        :param build_script: Path to the build script file.
        :raises ValueError: If any required variable is missing or incorrect.
        """
        checks = {
            #Check if MY_ROOT_DIR points to the correct folder
            r"export\s+MY_ROOT_DIR\s*=\s*/workdir/output/artifacts/__COAWST":
                "Wrong value for MY_ROOT_DIR in your build_coawst_script.\n"
                "MY_ROOT_DIR needs to be /workdir/output/artifacts/__COAWST",
            #Check if USE_MPI is on
            r"export\s+USE_MPI\s*=\s*on":
                "Wrong value for USE_MPI in your build_coawst_script.\n"
                "USE_MPI needs to be on",
            #Check if which_MPI is set to openmpi
            r"export\s+which_MPI\s*=\s*openmpi":
                "Wrong value for which_MPI in your build_coawst_script.\n"
                "which_MPI needs to be openmpi"
        }

        for regex, error_msg in checks.items():
            if not self._regex_exists_in_file(build_script, regex):
                raise ValueError(error_msg)

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            build_coawst_script: str,
            *,
            on: types.ComputationalResources,
            coawst_bin: str = "coawstM",
            init_commands: Optional[List[str]] = None,
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            build_coawst_script: Script used to build the COAWST executable.
            coawst_bin: Name of the COAWST binary to execute (coawstM by
            default).
            init_commands: List of helper commands to prepare things for your
                simulation. Used to copy files to and from
                `/workdir/output/artifacts/__COAWST`. It can also be used
                to run any helper function present within COAWST, like
                `scrip_coawst`.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """

        self._check_vcpus(n_vcpus, on)

        #only runs checks if we dont use remote assets
        if remote_assets is None:
            self._input_files_exist(input_dir=input_dir,
                                    sim_config_filename=sim_config_filename,
                                    build_coawst_script=build_coawst_script)

            build_script = os.path.join(input_dir, build_coawst_script)
            self._validate_build_script(build_script=str(build_script))

        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus
        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)

        # 34 selects dmpar for linux when compiling WRF
        compilation_command = Command(f"bash {build_coawst_script}", "34")

        commands = [
            #Copy COAWST source code to our input dir
            "cp -r /opt/COAWST /workdir/output/artifacts/__COAWST",
            "create_all_sim_links",
            #Compile COAWST
            compilation_command,
            Command(f"{coawst_bin} {sim_config_filename}",
                    mpi_config=mpi_config),
            "rm -r  __COAWST",
            "clean_all_sim_links"
        ]

        # Add init commands after building and before running the simulation
        if init_commands is not None:
            commands = commands[:1] + init_commands + commands[1:]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           resubmit_on_preemption=resubmit_on_preemption,
                           **kwargs)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineGroupUsageStatistics(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "idle_time",
            "total_time",
            "task_time",
        }

        class properties:

            @staticmethod
            def total_time() -> typing.Type['MachineGroupTimeStat']:
                return MachineGroupTimeStat

            @staticmethod
            def task_time() -> typing.Type['MachineGroupTimeStat']:
                return MachineGroupTimeStat

            @staticmethod
            def idle_time() -> typing.Type['MachineGroupTimeStat']:
                return MachineGroupTimeStat

            __annotations__ = {
                "total_time": total_time,
                "task_time": task_time,
                "idle_time": idle_time,
            }

    idle_time: 'MachineGroupTimeStat'
    total_time: 'MachineGroupTimeStat'
    task_time: 'MachineGroupTimeStat'

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["total_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["task_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["idle_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "total_time",
        "task_time",
        "idle_time",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["total_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["task_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["idle_time"]
    ) -> 'MachineGroupTimeStat':
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "total_time",
        "task_time",
        "idle_time",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        idle_time: 'MachineGroupTimeStat',
        total_time: 'MachineGroupTimeStat',
        task_time: 'MachineGroupTimeStat',
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MachineGroupUsageStatistics':
        return super().__new__(
            cls,
            *_args,
            idle_time=idle_time,
            total_time=total_time,
            task_time=task_time,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.machine_group_time_stat import MachineGroupTimeStat



from inductiva.client.paths.projects.get import ApiForget
from inductiva.client.paths.projects.post import ApiForpost
from inductiva.client.paths.projects.delete import ApiFordelete


class Projects(
        ApiForget,
        ApiForpost,
        ApiFordelete,
):
    pass



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TierFullInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "end_date",
            "capabilities",
            "quotas",
            "begin_date",
            "id",
            "label",
        }

        class properties:
            id = schemas.StrSchema
            label = schemas.StrSchema
            begin_date = schemas.DateTimeSchema
            end_date = schemas.DateTimeSchema

            class capabilities(schemas.ListSchema):

                class MetaOapg:

                    class items(schemas.DictSchema):

                        class MetaOapg:
                            additional_properties = schemas.StrSchema

                        def __getitem__(
                            self, name: typing.Union[
                                str,
                            ]
                        ) -> MetaOapg.additional_properties:
                            # dict_instance[name] accessor
                            return super().__getitem__(name)

                        def get_item_oapg(
                            self, name: typing.Union[
                                str,
                            ]
                        ) -> MetaOapg.additional_properties:
                            return super().get_item_oapg(name)

                        def __new__(
                            cls,
                            *_args: typing.Union[
                                dict,
                                frozendict.frozendict,
                            ],
                            _configuration: typing.Optional[
                                schemas.Configuration] = None,
                            **kwargs: typing.Union[
                                MetaOapg.additional_properties,
                                str,
                            ],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[
                        MetaOapg.items,
                        dict,
                        frozendict.frozendict,
                    ]], typing.List[typing.Union[
                        MetaOapg.items,
                        dict,
                        frozendict.frozendict,
                    ]]],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'capabilities':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)

            class quotas(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['TierQuotaDetail']:
                        return TierQuotaDetail

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['TierQuotaDetail'],
                                       typing.List['TierQuotaDetail']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'quotas':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'TierQuotaDetail':
                    return super().__getitem__(i)

            __annotations__ = {
                "id": id,
                "label": label,
                "begin_date": begin_date,
                "end_date": end_date,
                "capabilities": capabilities,
                "quotas": quotas,
            }

    end_date: MetaOapg.properties.end_date
    capabilities: MetaOapg.properties.capabilities
    quotas: MetaOapg.properties.quotas
    begin_date: MetaOapg.properties.begin_date
    id: MetaOapg.properties.id
    label: MetaOapg.properties.label

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["label"]
    ) -> MetaOapg.properties.label:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["begin_date"]
    ) -> MetaOapg.properties.begin_date:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["end_date"]
    ) -> MetaOapg.properties.end_date:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["capabilities"]
    ) -> MetaOapg.properties.capabilities:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["quotas"]
    ) -> MetaOapg.properties.quotas:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "label",
        "begin_date",
        "end_date",
        "capabilities",
        "quotas",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["label"]
    ) -> MetaOapg.properties.label:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["begin_date"]
    ) -> MetaOapg.properties.begin_date:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["end_date"]
    ) -> MetaOapg.properties.end_date:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["capabilities"]
    ) -> MetaOapg.properties.capabilities:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["quotas"]
    ) -> MetaOapg.properties.quotas:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "label",
        "begin_date",
        "end_date",
        "capabilities",
        "quotas",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        end_date: typing.Union[
            MetaOapg.properties.end_date,
            str,
            datetime,
        ],
        capabilities: typing.Union[
            MetaOapg.properties.capabilities,
            list,
            tuple,
        ],
        quotas: typing.Union[
            MetaOapg.properties.quotas,
            list,
            tuple,
        ],
        begin_date: typing.Union[
            MetaOapg.properties.begin_date,
            str,
            datetime,
        ],
        id: typing.Union[
            MetaOapg.properties.id,
            str,
        ],
        label: typing.Union[
            MetaOapg.properties.label,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TierFullInfo':
        return super().__new__(
            cls,
            *_args,
            end_date=end_date,
            capabilities=capabilities,
            quotas=quotas,
            begin_date=begin_date,
            id=id,
            label=label,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.tier_quota_detail import TierQuotaDetail



"""Project class"""
import contextvars
import datetime
import logging
import time
from typing import List, Optional, Union

import inductiva
from inductiva.client import ApiException
from inductiva.client import models
from inductiva.client.apis.paths.tasks import Tasks
from inductiva.client.apis.tags import projects_api
from inductiva.client.model.project import Project as ProjectModel
from inductiva.utils.format_utils import bytes_formatter, currency_formatter, timedelta_formatter

_logger = logging.getLogger(__name__)

_CURRENT_PROJECT = contextvars.ContextVar("current_project", default=None)


def get_current_project():
    """Get the current active project.
    
    Returns the currently active project in the calling thread.  If no
    project has been defined, the default project (`None`) will be
    returned.

    """
    return _CURRENT_PROJECT.get()


class ProjectInfo:
    """Stores info about the project."""

    def __init__(self, **attrs):
        self.task_by_status = {
            models.TaskStatusCode(attr): int(value)
            for attr, value in attrs.items()
        }


def get_projects():
    """Gets all the user's projects."""
    try:
        _logger.debug("Trying to get remote projects")
        api = projects_api.ProjectsApi(inductiva.api.get_client())
        response = api.get_user_projects()
    except ApiException as ex:
        _logger.error("Failed to get remote projects", exc_info=ex)
        raise ex

    # pylint: disable=protected-access
    return [Project._from_api_response(resp) for resp in response.body]


class Project:
    """Projects management class.
 
    The Project class allows tasks to be collected under the umbrella
    of a single container object. All tasks submitted in the context
    of an opened project will be associated with that project and can
    be retrieved altogether using the project as a filter. A project
    can be created and managed declaratively or using a context
    manager.
 
    Example usage:
 
    .. code-block:: python

        project = inductiva.projects.Project("test_project")
        project.start()
        simulator.run(...)  # Submitted to  `'test_project'`
        project.stop()
 
    This is equivalent to:
 
    .. code-block:: python

        with inductiva.projects.Project("test_project"):
            simulator.run(...)

    """

    def __init__(self, name: str, *, append: bool = False):
        """Initialize the Project instance.
        
        Args:
          name (str): The name of the project.
          append (bool): A flag indicating that new tasks can be appended to
              the project. If set to False (default), task submission will fail,
              even though the project might be opened.
        """
        self.append = append
        self._token = None
        self._list_of_tasks = None

        model = self._get_model(name)
        if model is None:
            model = self._create_model(name)
        self._update_from_api_response(model)

    @classmethod
    def _from_api_response(cls, model: ProjectModel) -> "Project":
        """Builds a `Project` object from an API response.

        Args:
          model: Info about the project. This should be fetched from the
          backend.

        """
        project = cls.__new__(Project)
        project.append = False
        project._token = None
        return project._update_from_api_response(model)

    @staticmethod
    def _get_model(name: str) -> ProjectModel:
        """Fetches the project info from the backend.

        Args:
          name: The name of the project

        Returns:
          Project models object with the info about the project.

        """
        try:
            _logger.debug("Trying to get remote project %s", name)
            api = projects_api.ProjectsApi(inductiva.api.get_client())
            response = api.get_project({"name": name})
            return response.body
        except ApiException as ex:
            if ex.status != 404:
                _logger.error("Failed to get remote project %s",
                              name,
                              exc_info=ex)
                raise ex
            _logger.debug("Project %s does not exist", name)
        return None

    @staticmethod
    def _create_model(name):
        """Creates a project with the given name on the backend."""
        try:
            _logger.debug("Creating remote project %s", name)
            api = projects_api.ProjectsApi(inductiva.api.get_client())
            response = api.create_project({"name": name})
            return response.body
        except ApiException as ex:
            _logger.error("Failed to create remote project %s",
                          name,
                          exc_info=ex)
            raise RuntimeError(f"Unable to create project {name}") from ex

    def _update_from_api_response(self, model: ProjectModel) -> "Project":
        """Updates itself with information fetched from the backend."""
        self._info = ProjectInfo(**model.get("task_status_overview"))
        self.created_at = model.get("created_at")
        self.num_tasks = int(model.get("num_tasks"))
        self._name = model.get("name")
        self.id = model.get("id")

        return self

    def open(self):
        """Open the project.

        Open the project and make it the active one in the calling
        thread. An opened project will ensure that calls to the
        `get_current_project` function will return this project.
        Consecutive calls to this method are idempotent.

        Raises:
            RuntimeError if another opened project exists, i.e.,
            `get_current_project` returns something other than None.

        """
        if not self.append:
            raise RuntimeError(
                "Trying to open a project with `append=False`.\n"
                "A Project can only be opened when instantiated with the"
                " `append=True` option.")

        _logger.debug("Opening project %s", self._name)
        current_project = get_current_project()

        if current_project is self:
            _logger.debug("Project is already opened.")
            return

        if current_project is not None:
            raise RuntimeError(
                "Trying to open a project when another is active.")

        self._token = _CURRENT_PROJECT.set(self)

    def close(self):
        """Close the project.

        Calls to the `get_current_project` will return `None` after
        the project is closed.  Consecutive calls to this method are
        idempotent.

        """
        if self._token is None:
            return

        _CURRENT_PROJECT.reset(self._token)
        self._token = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def opened(self) -> bool:
        """Checks if the project is open."""
        return self._token is not None

    @property
    def info(self) -> ProjectInfo:
        """Returns project info.

        It makes no requests to the backend. Instead it uses
        previously stored information. Therefore it can be
        outdated. For a method that requests the most recent data from
        the backend use: `get_info`.

        """
        return self._info

    def get_info(self) -> ProjectInfo:
        """Get project information.

        Get updated information on the project. This method executes a
        call to the backend to retrieve the most recent information
        about this project and stores it internally so that it becomes
        available through the `Project.info` property.  This method is
        suitable when up-to-date information is required.

        """
        name = self._name
        model = self._get_model(name)
        self._update_from_api_response(model)
        return self._info

    def _get_project_cost(self):
        """Returns the estimated project cost.

        The estimated project cost is the combination of the cost of all tasks.
        """
        list_of_tasks = self.list()
        total_cost = 0.0
        for task in list_of_tasks:
            total_cost += task.info.estimated_computation_cost
        return total_cost

    def __str__(self) -> str:
        project_cost = currency_formatter(self._get_project_cost())

        #Count status
        status_counts = {}
        total_files = 0
        total_size = 0

        list_of_tasks = self.list()
        total_duration = datetime.timedelta()
        running_tasks_warning = ""
        for task in list_of_tasks:
            status_counts[task.get_status()] = status_counts.get(
                task.get_status(), 0) + 1
            total_files += task.info.data_metrics.output_total_files.value
            total_size += task.info.data_metrics.output_size_bytes.value
            if task.info.start_time and task.info.end_time:
                start = datetime.datetime.fromisoformat(task.info.start_time)
                end = datetime.datetime.fromisoformat(task.info.end_time)
                duration_task = end - start
                total_duration += duration_task
            else:
                running_tasks_warning = (
                    "Warning: Some tasks may not be finished "
                    "yet. The values presented may change as a result.")

        # get start/end time
        start_project_time = min(
            datetime.datetime.fromisoformat(task.info.start_time)
            for task in list_of_tasks
            if task.info.start_time)
        end_project_time = max(
            datetime.datetime.fromisoformat(task.info.end_time)
            for task in list_of_tasks
            if task.info.end_time)

        duration = end_project_time - start_project_time

        total_size = bytes_formatter(total_size)
        duration = timedelta_formatter(duration)
        total_duration = timedelta_formatter(total_duration)

        status_report = ""
        for status, count in status_counts.items():
            status_report += f"{status}: {count}\n"

        return f"Project '{self._name}' with "\
               f"{self.num_tasks} tasks (id={self.id}).\n"\
               "\nTasks status:\n"\
               f"{status_report}"\
               f"\nTotal number of output files: {total_files}\n"\
               f"Total size of output: {total_size}\n"\
               f"\nProject duration: {duration}\n"\
               f"Project total simulated time: {total_duration}\n"\
               f"\nEstimated project cost: {project_cost}\n"\
               f"{running_tasks_warning}"

    def describe(self) -> str:
        """Generates a string description of the object

        Returns:
          str: A string description of the object. Includes project name,
            total number of tasks and the number of tasks by status.

        """
        header = str(self) + "\n"
        summary = "\n".join(
            f"  {k}: {v}" for k, v in self._info.task_by_status.items())
        return header + summary

    def get_tasks(
        self,
        last_n: int = -1,
        force_update=False,
        status: Optional[Union[str,
                               models.TaskStatusCode]] = None) -> List[Tasks]:
        """Get the the tasks of this project.

        Get the tasks that belong to this project,
        eventually filtered by status. By default, it will return all the tasks,
        irrespectively of their status.
        This method will only do a request to the back end if the list of
        tasks is None (never requested the list of tasks) or if `force_update`
        is passed as True.

        Args:
            last_n (int): The number of tasks with repect to the submission
                time to fectch. If `last_n<=0` we fetch all tasks submitted
                to the project.
            force_update (bool): Forces the request to the back end, even if we
                already have a list of tasks associated with this project.
            status: Status of the tasks to get. If `None`, tasks with any
                status will be returned.
        """
        logging.warning("Method `get_tasks` from projects is deprecated.\n"
                        "Use `list` instead.")
        return self.list(last_n=last_n,
                         force_update=force_update,
                         status=status)

    def list(
        self,
        last_n: int = -1,
        force_update=False,
        status: Optional[Union[str,
                               models.TaskStatusCode]] = None) -> List[Tasks]:
        """Get the the tasks of this project.

        Get the tasks that belong to this project,
        eventually filtered by status. By default, it will return all the tasks,
        irrespectively of their status.
        This method will only do a request to the back end if the list of
        tasks is None (never requested the list of tasks) or if `force_update`
        is passed as True.

        Args:
            last_n (int): The number of tasks with repect to the submission
                time to fectch. If `last_n<=0` we fetch all tasks submitted
                to the project.
            force_update (bool): Forces the request to the back end, even if we
                already have a list of tasks associated with this project.
            status: Status of the tasks to get. If `None`, tasks with any
                status will be returned.
        """
        if self._list_of_tasks is None or force_update is True:
            list_of_tasks = inductiva.tasks.get_tasks(last_n=last_n,
                                                      project=self,
                                                      status=status)

            self._list_of_tasks = list_of_tasks
        return self._list_of_tasks

    def wait(self):
        """ Wait for all the tasks in a project to complete."""
        all_tasks = self.list(force_update=True)
        print("Waiting for ALL tasks to finish")
        while (not all(x.is_terminal() for x in all_tasks)):
            finished = sum(x.is_terminal() for x in all_tasks)
            print(f"Finished: {finished} Total: {len(all_tasks)}", end="\r")
            time.sleep(5)
        print("All tasks in the project terminated.")

    def download_outputs(self):
        """ Downloads all the outputs for all the tasks in the project.
        
        All the files will be stored inside
        `inductiva_output/<project_name>/<task_id's>`.
        """
        list_of_tasks = self.list()

        for task in list_of_tasks:
            task.download_outputs(output_dir=f"{self.name}/{task.id}")

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __eq__(self, other) -> bool:
        return isinstance(other, Project) and \
            self.name == other.name and \
            self.id == other.id



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class OrganizationCostComponents(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "total",
            "infrastructure",
            "users",
        }

        class properties:

            class total(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['CostDetail']:
                        return CostDetail

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['CostDetail'],
                                       typing.List['CostDetail']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'total':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'CostDetail':
                    return super().__getitem__(i)

            class users(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['UserCostsDetails']:
                        return UserCostsDetails

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['UserCostsDetails'],
                                       typing.List['UserCostsDetails']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'users':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'UserCostsDetails':
                    return super().__getitem__(i)

            class infrastructure(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['CostDetail']:
                        return CostDetail

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['CostDetail'],
                                       typing.List['CostDetail']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'infrastructure':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'CostDetail':
                    return super().__getitem__(i)

            __annotations__ = {
                "total": total,
                "users": users,
                "infrastructure": infrastructure,
            }

    total: MetaOapg.properties.total
    infrastructure: MetaOapg.properties.infrastructure
    users: MetaOapg.properties.users

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["total"]
    ) -> MetaOapg.properties.total:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["users"]
    ) -> MetaOapg.properties.users:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["infrastructure"]
    ) -> MetaOapg.properties.infrastructure:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "total",
        "users",
        "infrastructure",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["total"]
    ) -> MetaOapg.properties.total:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["users"]
    ) -> MetaOapg.properties.users:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["infrastructure"]
    ) -> MetaOapg.properties.infrastructure:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "total",
        "users",
        "infrastructure",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        total: typing.Union[
            MetaOapg.properties.total,
            list,
            tuple,
        ],
        infrastructure: typing.Union[
            MetaOapg.properties.infrastructure,
            list,
            tuple,
        ],
        users: typing.Union[
            MetaOapg.properties.users,
            list,
            tuple,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'OrganizationCostComponents':
        return super().__new__(
            cls,
            *_args,
            total=total,
            infrastructure=infrastructure,
            users=users,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.cost_detail import CostDetail
from inductiva.client.model.user_costs_details import UserCostsDetails



"""OpenSees module of the API."""
from typing import List, Literal, Optional
import logging

from inductiva import types, tasks, simulators
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig

AVAILABLE_OPENSEES_INTERFACES = ["python", "tcl"]


@simulators.simulator.mpi_enabled
class OpenSees(simulators.Simulator):
    """Class to invoke a generic OpenSees simulation on the API."""

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 interface: Literal["python", "tcl"] = "python"):
        """Initialize the OpenSees simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            interface (str): The interface to use for interacting with the
                simulator. Can be either "python" (default) or "tcl".
        """
        if interface.lower() not in AVAILABLE_OPENSEES_INTERFACES:
            raise ValueError(
                f"Interface '{interface}' for OpenSees is not supported. "
                f"Available interfaces are: "
                f"{AVAILABLE_OPENSEES_INTERFACES}")
        self._interface = interface.lower()
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "opensees"

    @property
    def name(self):
        """Get the name of the simulator."""
        if self._interface == "python":
            return "OpenSeesPy"
        else:
            return "OpenSees"

    def _build_commands(self, sim_config_filename, mpi_config):
        """ Creates the list of commands based on the sim_config_filename and
        the provided mpi_config.
        """
        if self._interface == "python":
            return [
                Command(f"python {sim_config_filename}", mpi_config=mpi_config)
            ]
        return [
            Command(f"OpenSeesMP {sim_config_filename}", mpi_config=mpi_config)
        ]

    def _build_mpi_config(self, n_vcpus: int, use_hwthread: bool) -> MPIConfig:
        """ Creates an MPIConfig based on n_vcpus and use_hwthread.
        """
        return MPIConfig(version="4.1.6",
                         use_hwthread_cpus=use_hwthread,
                         **({
                             "np": n_vcpus
                         } if n_vcpus is not None else {}))

    def run(self,
            input_dir: Optional[str],
            *,
            on: types.ComputationalResources,
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            sim_config_filename: Optional[str] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            storage_dir: Directory for storing simulation results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            other arguments: See the documentation of the base class.
        """

        if self._version == "2.5.0":
            logging.warning(
                "Opensees version 2.5.0 does not support parallel execution."
                "Changing to `n_vcpus` to 1.")
            n_vcpus = 1
            if self._interface == "python":
                raise ValueError(
                    "Opensees version 2.5.0 does not support `python` as"
                    " an interface. Please use `interface='tcl'`.")

        self._check_vcpus(n_vcpus, on)

        mpi_config = self._build_mpi_config(n_vcpus, use_hwthread)

        if self._interface == "python":
            if n_vcpus or use_hwthread:
                logging.info("\nMPI flag detected (n_vcpus or use_hwthread)."
                             "\nThe simulation will run with mpirun "
                             "(Parallel Execution).")
                commands = self._build_commands(sim_config_filename, mpi_config)
            else:
                logging.info(
                    "\nNo MPI flag detected (n_vcpus or use_hwthread)."
                    "\nThe simulation will run sequentially with python.")
                commands = [f"python {sim_config_filename}"]
        else:
            commands = self._build_commands(sim_config_filename, mpi_config)

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskRequest(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "simulator",
            "resource_pool",
            "container_image",
        }

        class properties:
            simulator = schemas.StrSchema
            resource_pool = schemas.StrSchema
            container_image = schemas.StrSchema

            class simulator_name_alias(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'simulator_name_alias':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            storage_path_prefix = schemas.StrSchema

            class project(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'project':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class time_to_live_seconds(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'time_to_live_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            resubmit_on_preemption = schemas.BoolSchema

            class input_resources(schemas.ListSchema):

                class MetaOapg:
                    items = schemas.StrSchema

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[
                        MetaOapg.items,
                        str,
                    ]], typing.List[typing.Union[
                        MetaOapg.items,
                        str,
                    ]]],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'input_resources':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)

            stream_zip = schemas.BoolSchema

            class compress_with(
                    schemas.ComposedSchema,):

                class MetaOapg:

                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            CompressionMethod,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'compress_with':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "simulator": simulator,
                "resource_pool": resource_pool,
                "container_image": container_image,
                "simulator_name_alias": simulator_name_alias,
                "storage_path_prefix": storage_path_prefix,
                "project": project,
                "time_to_live_seconds": time_to_live_seconds,
                "resubmit_on_preemption": resubmit_on_preemption,
                "input_resources": input_resources,
                "stream_zip": stream_zip,
                "compress_with": compress_with,
            }

    simulator: MetaOapg.properties.simulator
    resource_pool: MetaOapg.properties.resource_pool
    container_image: MetaOapg.properties.container_image

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["simulator"]
    ) -> MetaOapg.properties.simulator:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["resource_pool"]
    ) -> MetaOapg.properties.resource_pool:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["container_image"]
    ) -> MetaOapg.properties.container_image:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["simulator_name_alias"]
    ) -> MetaOapg.properties.simulator_name_alias:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["storage_path_prefix"]
    ) -> MetaOapg.properties.storage_path_prefix:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["project"]
    ) -> MetaOapg.properties.project:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["time_to_live_seconds"]
    ) -> MetaOapg.properties.time_to_live_seconds:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["resubmit_on_preemption"]
    ) -> MetaOapg.properties.resubmit_on_preemption:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["input_resources"]
    ) -> MetaOapg.properties.input_resources:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["stream_zip"]
    ) -> MetaOapg.properties.stream_zip:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["compress_with"]
    ) -> MetaOapg.properties.compress_with:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "simulator",
        "resource_pool",
        "container_image",
        "simulator_name_alias",
        "storage_path_prefix",
        "project",
        "time_to_live_seconds",
        "resubmit_on_preemption",
        "input_resources",
        "stream_zip",
        "compress_with",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["simulator"]
    ) -> MetaOapg.properties.simulator:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["resource_pool"]
    ) -> MetaOapg.properties.resource_pool:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["container_image"]
    ) -> MetaOapg.properties.container_image:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["simulator_name_alias"]
    ) -> typing.Union[MetaOapg.properties.simulator_name_alias, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["storage_path_prefix"]
    ) -> typing.Union[MetaOapg.properties.storage_path_prefix, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["project"]
    ) -> typing.Union[MetaOapg.properties.project, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["time_to_live_seconds"]
    ) -> typing.Union[MetaOapg.properties.time_to_live_seconds, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["resubmit_on_preemption"]
    ) -> typing.Union[MetaOapg.properties.resubmit_on_preemption,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["input_resources"]
    ) -> typing.Union[MetaOapg.properties.input_resources, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["stream_zip"]
    ) -> typing.Union[MetaOapg.properties.stream_zip, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["compress_with"]
    ) -> typing.Union[MetaOapg.properties.compress_with, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "simulator",
        "resource_pool",
        "container_image",
        "simulator_name_alias",
        "storage_path_prefix",
        "project",
        "time_to_live_seconds",
        "resubmit_on_preemption",
        "input_resources",
        "stream_zip",
        "compress_with",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        simulator: typing.Union[
            MetaOapg.properties.simulator,
            str,
        ],
        resource_pool: typing.Union[
            MetaOapg.properties.resource_pool,
            str,
        ],
        container_image: typing.Union[
            MetaOapg.properties.container_image,
            str,
        ],
        simulator_name_alias: typing.Union[
            MetaOapg.properties.simulator_name_alias, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        storage_path_prefix: typing.Union[
            MetaOapg.properties.storage_path_prefix, str,
            schemas.Unset] = schemas.unset,
        project: typing.Union[MetaOapg.properties.project, dict,
                              frozendict.frozendict, str, date, datetime,
                              uuid.UUID, int, float, decimal.Decimal, bool,
                              None, list, tuple, bytes, io.FileIO,
                              io.BufferedReader, schemas.Unset] = schemas.unset,
        time_to_live_seconds: typing.Union[
            MetaOapg.properties.time_to_live_seconds, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        resubmit_on_preemption: typing.Union[
            MetaOapg.properties.resubmit_on_preemption, bool,
            schemas.Unset] = schemas.unset,
        input_resources: typing.Union[MetaOapg.properties.input_resources, list,
                                      tuple, schemas.Unset] = schemas.unset,
        stream_zip: typing.Union[MetaOapg.properties.stream_zip, bool,
                                 schemas.Unset] = schemas.unset,
        compress_with: typing.Union[MetaOapg.properties.compress_with, dict,
                                    frozendict.frozendict, str, date, datetime,
                                    uuid.UUID, int, float, decimal.Decimal,
                                    bool, None, list, tuple, bytes, io.FileIO,
                                    io.BufferedReader,
                                    schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TaskRequest':
        return super().__new__(
            cls,
            *_args,
            simulator=simulator,
            resource_pool=resource_pool,
            container_image=container_image,
            simulator_name_alias=simulator_name_alias,
            storage_path_prefix=storage_path_prefix,
            project=project,
            time_to_live_seconds=time_to_live_seconds,
            resubmit_on_preemption=resubmit_on_preemption,
            input_resources=input_resources,
            stream_zip=stream_zip,
            compress_with=compress_with,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.compression_method import CompressionMethod



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MultiPartExportOperation(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Body of the request to the multipart export files endpoint.
    """

    class MetaOapg:
        required = {
            "path",
            "parts_size",
            "complete_multipart_url",
            "upload_parts",
        }

        class properties:
            path = schemas.StrSchema
            parts_size = schemas.IntSchema

            class upload_parts(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['UploadPart']:
                        return UploadPart

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['UploadPart'],
                                       typing.List['UploadPart']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'upload_parts':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'UploadPart':
                    return super().__getitem__(i)

            class complete_multipart_url(schemas.StrSchema):
                pass

            __annotations__ = {
                "path": path,
                "parts_size": parts_size,
                "upload_parts": upload_parts,
                "complete_multipart_url": complete_multipart_url,
            }

    path: MetaOapg.properties.path
    parts_size: MetaOapg.properties.parts_size
    complete_multipart_url: MetaOapg.properties.complete_multipart_url
    upload_parts: MetaOapg.properties.upload_parts

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["path"]
    ) -> MetaOapg.properties.path:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["parts_size"]
    ) -> MetaOapg.properties.parts_size:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["upload_parts"]
    ) -> MetaOapg.properties.upload_parts:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["complete_multipart_url"]
    ) -> MetaOapg.properties.complete_multipart_url:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "path",
        "parts_size",
        "upload_parts",
        "complete_multipart_url",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["path"]
    ) -> MetaOapg.properties.path:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["parts_size"]
    ) -> MetaOapg.properties.parts_size:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["upload_parts"]
    ) -> MetaOapg.properties.upload_parts:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["complete_multipart_url"]
    ) -> MetaOapg.properties.complete_multipart_url:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "path",
        "parts_size",
        "upload_parts",
        "complete_multipart_url",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        path: typing.Union[
            MetaOapg.properties.path,
            str,
        ],
        parts_size: typing.Union[
            MetaOapg.properties.parts_size,
            decimal.Decimal,
            int,
        ],
        complete_multipart_url: typing.Union[
            MetaOapg.properties.complete_multipart_url,
            str,
        ],
        upload_parts: typing.Union[
            MetaOapg.properties.upload_parts,
            list,
            tuple,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MultiPartExportOperation':
        return super().__new__(
            cls,
            *_args,
            path=path,
            parts_size=parts_size,
            complete_multipart_url=complete_multipart_url,
            upload_parts=upload_parts,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.upload_part import UploadPart



"""
Converts a Docker image to an Apptainer .sif file via CLI.
"""

import argparse
import threading
import sys
import os
import tempfile
from typing import TextIO

from inductiva import constants

try:
    import docker
    from docker.errors import DockerException, ImageNotFound
except ImportError:
    docker = None


def generator_thread(container, fout: TextIO):
    """Stream logs from a container."""
    for line in container.logs(stream=True):
        print(line.decode("utf-8"), end="", file=fout)


def join_container_streams(*containers, fout: TextIO = sys.stdout):
    """Join multiple container log streams into one output."""
    container_threads = []
    for container in containers:
        thread = threading.Thread(target=generator_thread,
                                  args=(container, fout))
        thread.start()
        container_threads.append(thread)
    for thread in container_threads:
        thread.join()


def convert_image(args, fout: TextIO = sys.stdout):
    """
    Converts a Docker image to an Apptainer .sif file.
    
    The source image may be provided as:
      - a Docker Hub URL (e.g. docker://python:3.11-slim),
      - a local Docker image reference (by image name or ID),
      - or a path to a tar archive.
    
    The output file will be saved to the specified location.
    """
    if docker is None:
        print(
            "Docker Python API not installed. Please run "
            "'pip install inductiva[task-runner]' to install it.",
            file=fout,
        )
        return False

    try:
        client = docker.from_env()
    except DockerException as e:
        print(f"Failed to connect to Docker: {e}", file=fout)
        return False

    # Prepare output file path
    output = os.path.abspath(args.output)
    outdir = os.path.dirname(output)
    if outdir and not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)

    source = args.image
    conversion_source = None
    tmp_tar_path = None

    # Determine the conversion source
    if source.startswith("docker://"):
        # Use the docker:// URL directly.
        conversion_source = source
    elif os.path.exists(source):
        # Assume it's a tar file.
        conversion_source = f"docker-archive:///{os.path.abspath(source)}"
    else:
        # Assume it's a local Docker image reference.
        try:
            image_obj = client.images.get(source)
        except ImageNotFound:
            print(
                f"Local image '{source}' not found. "
                "\nThe image must be pulled or built in the system",
                file=fout)
            return False

        # Save the image to a temporary tar file.
        with tempfile.NamedTemporaryFile(suffix=".tar", delete=False) as tmp:
            tmp_tar_path = tmp.name
            with open(tmp_tar_path, "wb") as f:
                for chunk in image_obj.save(named=True):
                    f.write(chunk)
        conversion_source = f"docker-archive:///{tmp_tar_path}"

    mounts = []

    # If .tar or local docker image
    if tmp_tar_path:
        tar_host_path = os.path.abspath(tmp_tar_path)
        tar_host_dir = os.path.dirname(tar_host_path)
        tar_filename = os.path.basename(tar_host_path)
        tar_inside_container = f"/input/{tar_filename}"
        conversion_source = f"docker-archive://{tar_inside_container}"
        mounts.append(
            docker.types.Mount(target="/input",
                               source=tar_host_dir,
                               type="bind"))

    output_filename = os.path.basename(output)
    out_mount_path = "/output"
    mounts.append(
        docker.types.Mount(target=out_mount_path,
                           source=os.path.abspath(outdir),
                           type="bind"))

    command = [
        "build", f"{out_mount_path}/{output_filename}", conversion_source
    ]

    container = None
    try:
        container = client.containers.run(
            image=constants.TASK_RUNNER_IMAGE,
            name="apptainer-converter",
            entrypoint="/usr/bin/apptainer",
            command=command,
            mounts=mounts,
            platform="linux/amd64",
            detach=True,
            auto_remove=False,
        )

        print("Conversion started with container ID:",
              container.short_id,
              file=fout)
        join_container_streams(container, fout=fout)
        print(f"Conversion complete. Output saved at: {output}", file=fout)

    except KeyboardInterrupt:
        print("\n⚠️ Conversion interrupted by user. Cleaning up...", file=fout)
        if container:
            try:
                container.kill()
                container.remove(force=True)
                print("🧹 Container stopped and removed.", file=fout)
            except Exception as cleanup_err:  # pylint: disable=broad-exception-caught
                print(f"⚠️ Failed to clean up container: {cleanup_err}",
                      file=fout)
        return False

    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"Error during conversion: {e}", file=fout)
        return False

    # Clean up the temporary tar file if one was created.
    if tmp_tar_path and os.path.exists(tmp_tar_path):
        try:
            os.remove(tmp_tar_path)
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(
                f"Warning: could not remove temporary file {tmp_tar_path}: {e}",
                file=fout)
            return False

    return True


def register(parser):
    """Register the convert command."""
    subparser = parser.add_parser(
        "convert",
        help="Convert a Docker image to an Apptainer .sif file.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparser.description = (
        "The `inductiva containers convert` command converts a Docker "
        "image—specified as a Docker Hub URL,\n a local Docker image reference,"
        " or a .tar archive—into an Apptainer .sif file using\n the Apptainer.")
    subparser.add_argument(
        "image",
        type=str,
        help=("Docker image reference. A docker:// URL, a local image name/ID, "
              "or a path to a .tar file."),
    )
    subparser.add_argument(
        "output",
        type=str,
        help="Path to save the resulting .sif file.",
    )
    subparser.set_defaults(func=convert_image)



"""Base class for low-level simulators."""
from typing import List, Literal, Optional
from abc import ABC
import logging
import os
import re

import pathlib

from inductiva import types, tasks, resources
from .methods import list_available_images
from inductiva import commands


def mpi_enabled(cls):
    """Class decorator that adds MPICluster to the supported resources.
    """
    supports = getattr(cls, "_supported_resources", set())
    cls._supported_resources = supports | {resources.MPICluster}  # pylint: disable=protected-access

    return cls


class Simulator(ABC):
    """Base simulator class."""

    _supported_resources = {
        resources.MachineGroup, resources.ElasticMachineGroup
    }
    _logger = logging.getLogger(__name__)

    _supported_device_values = ["gpu", "cpu", "auto"]

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): Specifies whether to use the CPU or GPU version of
                the Docker image. If `auto` is picked we will pick the
                appropriate device based on the machine used to run the
                simulation, meaning, if the machine has a GPU we will pick
                GPU, otherwise we pick CPU.
        """
        if version is not None and not isinstance(version, str):
            raise ValueError("Version must be a string or None.")

        if  device is not None and \
            device.lower() not in self._supported_device_values:
            raise ValueError("Wrong value for `device`. Supported "
                             "values are `gpu`, `cpu` or `auto`.")
        self.simulator = ""
        self.simulator_name_alias = None
        self._version = version
        self._use_dev = bool(use_dev)
        self._image_uri = self._get_image_uri()
        self.container_image = self._image_uri
        self._logger.info("")
        self._device = device.lower() if device else None

    @property
    def version(self):
        """Get the version of the simulator."""
        return self._version

    @property
    def use_dev(self):
        """Get whether the development version of the simulator is used."""
        return self._use_dev

    @property
    def name(self):
        """Get the name of the simulator."""
        return self.__class__.__name__

    @property
    def image_uri(self):
        """Get the image URI for this simulator."""
        return self._image_uri

    def _input_files_exist(self, input_dir, **kwargs):
        """
        Checks if all the files in kwargs are present in the input_dir.
        """
        missing_files = []

        for _, file_path in kwargs.items():
            # Get the full file path by joining the input directory with the
            # file path in kwargs
            full_file_path = os.path.join(input_dir, file_path)

            # Check if the file exists
            if not os.path.isfile(full_file_path):
                missing_files.append(file_path)

        if missing_files:
            missing_files_str = ", ".join(missing_files)
            raise FileNotFoundError(
                "The following files are missing from your input directory:\n"
                f"{missing_files_str}")

    def _regex_exists_in_file(self, file_path: str, pattern: str) -> bool:
        """
        Check if a given regular expression exists in a file.
        
        :param file_path: Path to the file.
        :param pattern: Regular expression pattern to search for.
        :return: True if the pattern exists, False otherwise.
        """
        with open(file_path, "r", encoding="utf-8") as file:
            for line in file:
                if re.search(pattern, line):
                    return True
        return False

    def _check_vcpus(
        self,
        n_vcpus: Optional[int],
        machine: types.ComputationalResources,
    ) -> bool:
        """ Checks if the machine supports the number of n_vcpus passed."""
        if n_vcpus is not None and n_vcpus > machine.n_vcpus.total:
            raise ValueError(
                "The number of virtual cpus asked surpasses the"
                " available virtual cpus for the selected resource.")

    def _get_image_uri(self):
        """Get the appropriate image name for this simulator."""

        img_type = "development" if self._use_dev else "production"
        sim_name = self.name
        name = sim_name.lower()

        available = list_available_images()
        listing = available.get(img_type, {}).get(name, [])

        self._supported_versions = {
            element.split("_")[0] for element in listing
        }

        self._supported_versions_with_suffixes = listing

        if self._version is None:
            # kutu does not have a specific tag for the latest version
            # this hack is a workaround to get that version, but it is prone
            # to errors.
            self._version = max(self._supported_versions)

        if self._version not in self._supported_versions:
            raise ValueError(
                f"Version {self.version} is not available for simulator {name}."
                f" Available versions are: {self._supported_versions}.")
        self._logger.info("■ Using %s image of %s version %s", img_type,
                          sim_name, self.version)

        return f"docker://inductiva/kutu:{name}_v{self._version}"

    @classmethod
    def get_supported_resources(cls):
        """Get the supported computational resources for this simulator."""
        return tuple(cls._supported_resources)

    def _setup_input_dir(self, input_dir: str):
        """Setup the simulator input directory."""
        input_dir_path = pathlib.Path(input_dir)
        if not input_dir_path.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")
        return input_dir_path

    def _validate_input_files(self, input_dir, remote_assets):
        if input_dir == "":
            raise ValueError(
                "input_dir cannot be an empty string. Use None instead.")

        if not input_dir and not remote_assets:
            raise ValueError(
                "Either input_dir or remote_assets must be provided.")

    def _get_version_suffixes(self,
                              resource: types.ComputationalResources) -> str:
        """
        Gets the suffixes based on the resource used for the simulation.

        If the resource has a GPU we will prioritize that.

        :param resource: The resource used for the simulation.
        :return: A string of suffixes.
        """
        dev_suffix = "_dev" if self._use_dev else ""
        gpu_suffix = "_gpu" if resource.has_gpu() else ""

        # Overwrites the gpu suffix if the user passed a specific
        # device
        if self._device == "gpu":
            gpu_suffix = "_gpu"
        elif self._device == "cpu":
            gpu_suffix = ""

        # Add new suffixes here
        suffix = f"{gpu_suffix}"

        if self._device and self._device == "gpu" and not resource.has_gpu():
            raise ValueError("The selected machine needs to have a GPU.\n"
                             "If available, either pick device=`cpu` or pick"
                             " a different machine.")

        suffix = f"{suffix}{dev_suffix}"
        return suffix

    def get_simulator_image_based_on_resource(
            self, resource: types.ComputationalResources):
        """
        Get the simulator image for the simulation, based on the resourced used.

        This method takes in consideration the specified device for the
        simulation. If device=`auto` we will pick the appropriate device based
        on the machine.

        :param resource: The computational resource to use for the simulation.
        :return: The simulator image based on the resource used.
        """
        suffixes = ""

        if self._device:
            suffixes = self._get_version_suffixes(resource)

        return f"{self._image_uri}{suffixes}"

    def run(
        self,
        input_dir: Optional[str],
        *_args,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        remote_assets: Optional[List[str]] = None,
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            on: The computational resource to launch the simulation in. If None
                the simulation is launched in a machine of the default pool.
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            storage_dir: Parent directory for storing simulation
                               results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        self._validate_input_files(input_dir, remote_assets)

        input_dir_path = self._setup_input_dir(input_dir) if input_dir else None

        if on is None:
            raise ValueError(
                "A default computational resource is no longer "
                "provided. Please specify a computational resource. "
                "Check https://docs.inductiva.ai/en/latest/how_to/"
                "run-parallel_simulations.html "
                "to learn how to create your own computational resource.")

        self.validate_computational_resources(on)

        if "commands" in kwargs:
            cmds = commands.Command.commands_to_dicts(kwargs["commands"])
            kwargs["commands"] = cmds

        # Get the user-specified image name. If not specified,
        # use the default image name for the current simulator
        self._image_uri = kwargs.pop("container_image", self._image_uri)

        # CustomImage does not use suffixes. We can the image as is
        if self.__class__.__name__ != "CustomImage":
            self._image_uri = self.get_simulator_image_based_on_resource(on)

        if on.has_gpu() and "_gpu" not in self._image_uri:
            logging.warning("Attention: The machine you selected has a GPU, but"
                            " the simulator you picked will run on the CPU "
                            "only.\n")

        return tasks.run_simulation(
            self.simulator,
            input_dir_path,
            simulator_obj=self,
            storage_dir=storage_dir,
            machine_group=on,
            container_image=self._image_uri,
            resubmit_on_preemption=resubmit_on_preemption,
            remote_assets=remote_assets,
            simulator_name_alias=self.simulator_name_alias,
            **kwargs,
        )

    def validate_computational_resources(self, resource):
        """Validate the computational resources passed to the run method.

        Args:
            resource: The computational resource to validate.
            valid_resources: The valid computational resources for the simulator
        """

        if not isinstance(resource, tuple(self._supported_resources)):
            raise ValueError(
                "The computational resource is invalid for this simulator. "
                f"Expected one of {self._supported_resources} but got "
                f"{resource}.")



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class FileInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "name",
        }

        class properties:
            name = schemas.StrSchema

            class size(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'size':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class compressed_size(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'compressed_size':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class range_start(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'range_start':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class creation_time(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'creation_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class compress_type(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'compress_type':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "name": name,
                "size": size,
                "compressed_size": compressed_size,
                "range_start": range_start,
                "creation_time": creation_time,
                "compress_type": compress_type,
            }

    name: MetaOapg.properties.name

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["size"]
    ) -> MetaOapg.properties.size:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["compressed_size"]
    ) -> MetaOapg.properties.compressed_size:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["range_start"]
    ) -> MetaOapg.properties.range_start:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["creation_time"]
    ) -> MetaOapg.properties.creation_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["compress_type"]
    ) -> MetaOapg.properties.compress_type:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "size",
        "compressed_size",
        "range_start",
        "creation_time",
        "compress_type",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["size"]
    ) -> typing.Union[MetaOapg.properties.size, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["compressed_size"]
    ) -> typing.Union[MetaOapg.properties.compressed_size, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["range_start"]
    ) -> typing.Union[MetaOapg.properties.range_start, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["creation_time"]
    ) -> typing.Union[MetaOapg.properties.creation_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["compress_type"]
    ) -> typing.Union[MetaOapg.properties.compress_type, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "size",
        "compressed_size",
        "range_start",
        "creation_time",
        "compress_type",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        size: typing.Union[MetaOapg.properties.size, dict,
                           frozendict.frozendict, str, date, datetime,
                           uuid.UUID, int, float, decimal.Decimal, bool, None,
                           list, tuple, bytes, io.FileIO, io.BufferedReader,
                           schemas.Unset] = schemas.unset,
        compressed_size: typing.Union[MetaOapg.properties.compressed_size, dict,
                                      frozendict.frozendict, str, date,
                                      datetime, uuid.UUID, int, float,
                                      decimal.Decimal, bool, None, list, tuple,
                                      bytes, io.FileIO, io.BufferedReader,
                                      schemas.Unset] = schemas.unset,
        range_start: typing.Union[MetaOapg.properties.range_start, dict,
                                  frozendict.frozendict, str, date, datetime,
                                  uuid.UUID, int, float, decimal.Decimal, bool,
                                  None, list, tuple, bytes, io.FileIO,
                                  io.BufferedReader,
                                  schemas.Unset] = schemas.unset,
        creation_time: typing.Union[MetaOapg.properties.creation_time, dict,
                                    frozendict.frozendict, str, date, datetime,
                                    uuid.UUID, int, float, decimal.Decimal,
                                    bool, None, list, tuple, bytes, io.FileIO,
                                    io.BufferedReader,
                                    schemas.Unset] = schemas.unset,
        compress_type: typing.Union[MetaOapg.properties.compress_type, dict,
                                    frozendict.frozendict, str, date, datetime,
                                    uuid.UUID, int, float, decimal.Decimal,
                                    bool, None, list, tuple, bytes, io.FileIO,
                                    io.BufferedReader,
                                    schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'FileInfo':
        return super().__new__(
            cls,
            *_args,
            name=name,
            size=size,
            compressed_size=compressed_size,
            range_start=range_start,
            creation_time=creation_time,
            compress_type=compress_type,
            _configuration=_configuration,
            **kwargs,
        )



"""SplisHSPlasH simulator module of the API."""
from typing import List, Optional

from inductiva import types, tasks, simulators


class SplishSplash(simulators.Simulator):
    """Class to invoke a generic SPlisHSPlasH simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the SPlisHSplasH simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "splishsplash"

    def run(
        self,
        input_dir: Optional[str],
        sim_config_filename: str,
        *,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        remote_assets: Optional[List[str]] = None,
        **kwargs,
    ) -> tasks.Task:
        """Run the SPlisHSPlasH simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Directory for storing simulation results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        Returns:
            Task object representing the simulation task.
        """

        commands = [
            "cp /SPlisHSPlasH_CPU/bin/SPHSimulator .",
            f"./SPHSimulator {sim_config_filename} --no-gui --output-dir .",
            "rm SPHSimulator"
        ]
        return super().run(
            input_dir,
            commands=commands,
            storage_dir=storage_dir,
            on=on,
            resubmit_on_preemption=resubmit_on_preemption,
            remote_assets=remote_assets,
            **kwargs,
        )



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    TASKS_SUBMIT = "/tasks/submit"
    TASKS_TASK_ID_INPUT_UPLOADED = "/tasks/{task_id}/input_uploaded"
    TASKS_TASK_ID = "/tasks/{task_id}"
    TASKS = "/tasks"
    TASKS_TASK_ID_STATUS = "/tasks/{task_id}/status"
    TASKS_TASK_ID_KILL = "/tasks/{task_id}/kill"
    TASKS_TASK_ID_REGISTER = "/tasks/{task_id}/register"
    TASKS_TASK_ID_OFFER = "/tasks/{task_id}/offer"
    TASKS_TASK_ID_MESSAGE = "/tasks/{task_id}/message"
    ADMIN_USERS = "/admin/users"
    ADMIN_USERS_EMAIL_TERMS_AND_CONDITIONS = "/admin/users/{email}/terms_and_conditions"
    ADMIN_USERS_USERNAME_ORGANIZATION = "/admin/users/{username}/organization"
    ADMIN_USERS_USERNAME_CREDITS = "/admin/users/{username}/credits"
    ADMIN_USER_EMAILS = "/admin/user_emails"
    ADMIN_USER_BUCKETS = "/admin/user_buckets"
    ADMIN_USERS_EMAIL_API_KEY = "/admin/users/{email}/api_key"
    ADMIN_USERS_EMAIL = "/admin/users/{email}"
    ADMIN_USERS_USERNAME_STORAGE_SIZE_FS = "/admin/users/{username}/storage_size_fs"
    ADMIN_USERS_USERNAME_STORAGE_SIZE = "/admin/users/{username}/storage_size"
    ADMIN_USERS_USERNAME_TASKS = "/admin/users/{username}/tasks"
    ADMIN_USERS_USERNAME_CAPABILITIES = "/admin/users/{username}/capabilities"
    ADMIN_GROUPS = "/admin/groups"
    ADMIN_GROUPS_ACTIVE = "/admin/groups/active"
    ADMIN_GROUPS_ID = "/admin/groups/{id}"
    ADMIN_ACTIVE_TASKS = "/admin/active_tasks"
    ADMIN_GROUPS_MACHINE_GROUP_ID_TERMINATE = "/admin/groups/{machine_group_id}/terminate"
    ADMIN_ORGANIZATIONS = "/admin/organizations"
    ADMIN_ORGANIZATIONS_ORGANIZATION_ID = "/admin/organizations/{organization_id}"
    ADMIN_ORGANIZATIONS_COSTS = "/admin/organizations/costs"
    ADMIN_TIERS = "/admin/tiers"
    ADMIN_TERMINATE_MACHINE_GROUPS_CREDITS_EXHAUSTED = "/admin/terminate_machine_groups_credits_exhausted"
    ADMIN_IMPORT_PROVIDER_COSTS = "/admin/import_provider_costs"
    ADMIN_MACHINE_ID_EVENT = "/admin/{machine_id}/event"
    ADMIN_USERS_USERNAME_COSTS_FEE_PERCENTAGE = "/admin/users/{username}/costs_fee_percentage"
    ADMIN_ORGANIZATIONS_ORGANIZATION_ID_COSTS_FEE_PERCENTAGE = "/admin/organizations/{organization_id}/costs_fee_percentage"
    ADMIN_TOPUPS = "/admin/top-ups"
    ADMIN_TASKS = "/admin/tasks"
    ADMIN_USERS_EMAIL_STRIPE_CUSTOMER_ID = "/admin/users/{email}/stripe_customer_id"
    ADMIN_FEATUREFLAGS_NAME = "/admin/feature-flags/{name}"
    ADMIN_FEATUREFLAGS_ = "/admin/feature-flags/"
    TASKRUNNER_REGISTER = "/task-runner/register"
    TASKRUNNER_MACHINE_ID = "/task-runner/{machine_id}"
    TASKRUNNER_MACHINE_ID_TASK = "/task-runner/{machine_id}/task"
    TASKRUNNER_MACHINE_ID_TASK_TASK_ID_MESSAGE = "/task-runner/{machine_id}/task/{task_id}/message"
    TASKRUNNER_MACHINE_ID_TASK_TASK_ID_MESSAGE_UNBLOCK = "/task-runner/{machine_id}/task/{task_id}/message/unblock"
    TASKRUNNER_MACHINE_ID_EVENT = "/task-runner/{machine_id}/event"
    TASKRUNNER_MACHINE_ID_TASK_TASK_ID_OPERATION = "/task-runner/{machine_id}/task/{task_id}/operation"
    TASKRUNNER_MACHINE_ID_TASK_TASK_ID_OPERATION_OPERATION_ID_DONE = "/task-runner/{machine_id}/task/{task_id}/operation/{operation_id}/done"
    TASKRUNNER_MACHINE_ID_TASK_TASK_ID_METRIC = "/task-runner/{machine_id}/task/{task_id}/metric"
    TASKRUNNER_MACHINE_ID_RESIZE_DISK = "/task-runner/{machine_id}/resize_disk"
    TASKRUNNER_MACHINE_ID_RESIZE_DISK_DONE = "/task-runner/{machine_id}/resize_disk_done"
    COMPUTE_GROUP = "/compute/group"
    COMPUTE_GROUP_START = "/compute/group/start"
    COMPUTE_PRICE = "/compute/price"
    COMPUTE_GROUPS = "/compute/groups"
    COMPUTE_GROUPS_HISTORY = "/compute/groups/history"
    COMPUTE_MACHINE_TYPES = "/compute/machine_types"
    COMPUTE_GROUP_NAME = "/compute/group/{name}"
    COMPUTE_GROUP_MACHINE_GROUP_ID_SHARING = "/compute/group/{machine_group_id}/sharing"
    STORAGE_SIZE = "/storage/size"
    STORAGE_COST = "/storage/cost"
    STORAGE_CONTENTS = "/storage/contents"
    STORAGE_ZIPCONTENTS = "/storage/zip-contents"
    STORAGE_SIGNEDURLS = "/storage/signed-urls"
    STORAGE_INPUT_NOTIFY = "/storage/input_notify"
    STORAGE_INPUT_REMOTE = "/storage/input_remote"
    STORAGE_ = "/storage/"
    STORAGE_COPY = "/storage/copy"
    STORAGE_OPERATIONS_OPERATION_ID = "/storage/operations/{operation_id}"
    STORAGE_OPERATIONS = "/storage/operations"
    STORAGE_EXPORT_MULTIPART = "/storage/export_multipart"
    STORAGE_UPDATE_OPERATION_STATUS = "/storage/update_operation_status"
    VERSION = "/version"
    VERSIONCHECK = "/version-check"
    USERS_QUOTAS = "/users/quotas"
    USERS_INFO = "/users/info"
    USERS_CAPABILITIES = "/users/capabilities"
    USERS_COSTS = "/users/costs"
    USERS_ORGANIZATION_COSTS = "/users/organization/costs"
    USERS_TOPUPS = "/users/top-ups"
    PROJECTS = "/projects"
    PROJECTS_NAME = "/projects/{name}"
    PROJECTS_NAME_TASK_TASK_ID_ADD = "/projects/{name}/task/{task_id}/add"
    PROJECTS_NAME_TASK_TASK_ID_REMOVE = "/projects/{name}/task/{task_id}/remove"
    METRICS_USERS_USERNAME_ACTIVITY = "/metrics/users/{username}/activity"
    METRICS_USERS_USERNAME_COST_OVER_TIME = "/metrics/users/{username}/cost_over_time"
    METRICS_USERS_USERNAME_TASK_STATUS_OVERVIEW = "/metrics/users/{username}/task_status_overview"
    METRICS_USERS_USERNAME_COMPUTATION_TIME_TREND = "/metrics/users/{username}/computation_time_trend"
    METRICS_USERS_USERNAME_TASKS_OVERVIEW = "/metrics/users/{username}/tasks_overview"
    METRICS_USERS_USERNAME_MOST_USED_MACHINE_TYPES = "/metrics/users/{username}/most_used_machine_types"
    METRICS_USERS_USERNAME_MOST_USED_SIMULATORS_OVERVIEW = "/metrics/users/{username}/most_used_simulators_overview"
    METRICS_USAGE_STATISTICS = "/metrics/usage_statistics"



"""DualSPHysics simulator module of the API."""

from typing import List, Literal, Optional

from inductiva import types, tasks, simulators


class DualSPHysics(simulators.Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the DualSPHysics simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): Select between CPU or GPU for running the simulation.
                Default is "auto", which will auto-detect if the machine has a
                GPU and use if it available, otherwise use the CPU.
        """
        super().__init__(version=version, use_dev=use_dev, device=device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "dualsphysics"

    def run(
        self,
        input_dir: Optional[str],
        shell_script: str,
        *,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        remote_assets: Optional[List[str]] = None,
        **kwargs,
    ) -> tasks.Task:
        """Executes a DualSPHysics simulation.

        Args:
            input_dir: Directory with simulation input files.
            shell_script: Path to the shell script to run the simulation.
            on: The computational resource to launch the simulation on.
            storage_dir: Directory for storing results.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.

        Returns:
            tasks.Task: An object representing the simulation task.
        """
        commands = [f"bash {shell_script}"]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)
